/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * pikasymmetry.h
 * Copyright (C) 2015 Jehan <jehan@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __PIKA_SYMMETRY_H__
#define __PIKA_SYMMETRY_H__


#include "pikaobject.h"


/* shift one more than PIKA_CONFIG_PARAM_DONT_COMPARE */
#define PIKA_SYMMETRY_PARAM_GUI (1 << (7 + G_PARAM_USER_SHIFT))


#define PIKA_TYPE_SYMMETRY            (pika_symmetry_get_type ())
#define PIKA_SYMMETRY(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), PIKA_TYPE_SYMMETRY, PikaSymmetry))
#define PIKA_SYMMETRY_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), PIKA_TYPE_SYMMETRY, PikaSymmetryClass))
#define PIKA_IS_SYMMETRY(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), PIKA_TYPE_SYMMETRY))
#define PIKA_IS_SYMMETRY_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), PIKA_TYPE_SYMMETRY))
#define PIKA_SYMMETRY_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), PIKA_TYPE_SYMMETRY, PikaSymmetryClass))


typedef struct _PikaSymmetryClass   PikaSymmetryClass;

struct _PikaSymmetry
{
  PikaObject    parent_instance;

  PikaImage    *image;
  PikaDrawable *drawable;
  PikaCoords   *origin;
  gboolean      active;
  gint          version;

  GList        *strokes;
  gboolean      stateful;
};

struct _PikaSymmetryClass
{
  PikaObjectClass  parent_class;

  const gchar * label;

  /* Virtual functions */
  void       (* update_strokes)             (PikaSymmetry       *symmetry,
                                             PikaDrawable       *drawable,
                                             PikaCoords         *origin);
  void       (* get_transform)              (PikaSymmetry       *symmetry,
                                             gint                stroke,
                                             gdouble            *angle,
                                             gboolean           *reflect);
  void       (* active_changed)             (PikaSymmetry       *symmetry);

  gboolean   (* update_version)             (PikaSymmetry       *symmetry);
};


GType          pika_symmetry_get_type       (void) G_GNUC_CONST;

void           pika_symmetry_set_stateful   (PikaSymmetry       *symmetry,
                                             gboolean            stateful);
void           pika_symmetry_set_origin     (PikaSymmetry       *symmetry,
                                             PikaDrawable       *drawable,
                                             PikaCoords         *origin);
void           pika_symmetry_clear_origin   (PikaSymmetry       *symmetry);

PikaCoords   * pika_symmetry_get_origin     (PikaSymmetry       *symmetry);
gint           pika_symmetry_get_size       (PikaSymmetry       *symmetry);
PikaCoords   * pika_symmetry_get_coords     (PikaSymmetry       *symmetry,
                                             gint                stroke);
void           pika_symmetry_get_transform  (PikaSymmetry       *symmetry,
                                             gint                stroke,
                                             gdouble            *angle,
                                             gboolean           *reflect);
void           pika_symmetry_get_matrix     (PikaSymmetry       *symmetry,
                                             gint                stroke,
                                             PikaMatrix3        *matrix);
GeglNode     * pika_symmetry_get_operation  (PikaSymmetry       *symmetry,
                                             gint                stroke);

gchar        * pika_symmetry_parasite_name  (GType               type);
PikaParasite * pika_symmetry_to_parasite    (const PikaSymmetry *symmetry);
PikaSymmetry * pika_symmetry_from_parasite  (const PikaParasite *parasite,
                                             PikaImage          *image,
                                             GType               type);


#endif  /*  __PIKA_SYMMETRY_H__  */
